! LOGSTEP Module: batch_sep_dc.F90
!
! * This program is written by Yusuke HIROTA.
!
! # License (MIT License)
! Copyright 2018 RIKEN.
!
! Permission is hereby granted, free of charge, to any person obtaining
! a copy of this software and associated documentation files (the
! "Software"), to deal in the Software without restriction, including
! without limitation the rights to use, copy, modify, merge, publish,
! distribute, sublicense, and/or sell copies of the Software, and to
! permit persons to whom the Software is furnished to do so, subject to
! the following conditions:
!
! The above copyright notice and this permission notice shall be
! included in all copies or substantial portions of the Software.
!
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
! EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
! MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
! IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
! CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
! TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
! SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
!
module batch_sep_dc_timing
  implicit none
  integer, parameter :: batch_sep_dc_timing_max_level = 6
  type :: batch_sep_dc_timing_list
    real(kind(0.0d0)) :: t_dstedc_batch = 0.0d0
    real(kind(0.0d0)) :: t_dlaed0_batch = 0.0d0
    real(kind(0.0d0)) :: t_dlaed0b_batch = 0.0d0
    integer :: current_level = -999 ! Initialized by an arbitrary negative integer
    real(kind(0.0d0)) :: t_dlaed1a_batch(batch_sep_dc_timing_max_level) = 0.0d0
    real(kind(0.0d0)) :: t_dlaed1b_batch(batch_sep_dc_timing_max_level) = 0.0d0
    real(kind(0.0d0)) :: t_dlaed2(batch_sep_dc_timing_max_level) = 0.0d0
    real(kind(0.0d0)) :: t_dlaed3_batch(batch_sep_dc_timing_max_level) = 0.0d0
    real(kind(0.0d0)) :: t_batch_dgemm(batch_sep_dc_timing_max_level) = 0.0d0
    real(kind(0.0d0)) :: t_dlaed4(batch_sep_dc_timing_max_level) = 0.0d0
    real(kind(0.0d0)) :: t_general(batch_sep_dc_timing_max_level) = 0.0d0
  endtype
contains
  subroutine batch_sep_dc_timing_reset(timer)
    implicit none
    type(batch_sep_dc_timing_list), intent(out) :: timer
    timer%t_dstedc_batch = 0.0d0
    timer%t_dlaed0_batch = 0.0d0
    timer%t_dlaed0b_batch = 0.0d0
    timer%t_dlaed1a_batch(:) = 0.0d0
    timer%t_dlaed1b_batch(:) = 0.0d0
    timer%t_dlaed2(:) = 0.0d0
    timer%t_dlaed3_batch(:) = 0.0d0
    timer%t_batch_dgemm(:) = 0.0d0
    timer%t_dlaed4(:) = 0.0d0
    timer%t_general(:) = 0.0d0
  end subroutine
end

module batch_dgemm
  use omp_lib
  implicit none
  type :: batch_dgemm_parameter
    integer :: group_count
    integer, allocatable :: group_size(:)
    character, allocatable :: transa(:), transb(:) 
    real(kind(0.0d0)), allocatable :: alpha(:), beta(:)
    integer, allocatable :: m(:), n(:), k(:)
    integer, allocatable :: lda(:), ldb(:), ldc(:)
    integer(kind = C_SIZE_T), allocatable :: ap(:), bp(:), cp(:)
  endtype
contains
  subroutine batch_dgemm_parameter_allocate(b)
    ! group_size(:) should be allocated before this subroutine.
    implicit none
    type(batch_dgemm_parameter), intent(inout) :: b
    integer :: total_batch_count
    allocate(b%transa(b%group_count), b%transb(b%group_count))
    allocate(b%alpha(b%group_count), b%beta(b%group_count))
    allocate(b%m(b%group_count), b%n(b%group_count), b%k(b%group_count))
    allocate(b%lda(b%group_count), b%ldb(b%group_count), b%ldc(b%group_count))
    total_batch_count = sum(b%group_size(1 : b%group_count))
    allocate(b%ap(total_batch_count), b%bp(total_batch_count), b%cp(total_batch_count))
  end subroutine
  subroutine batch_dgemm_parameter_deallocate(b) ! group_size(:) is not freed in this function.
    implicit none
    type(batch_dgemm_parameter), intent(inout) :: b
    deallocate(b%transa, b%transb)
    deallocate(b%alpha, b%beta)
    deallocate(b%m, b%n, b%k)
    deallocate(b%lda, b%ldb, b%ldc)
    deallocate(b%ap, b%bp, b%cp)
  end subroutine
  subroutine batch_dgemm_wrapper(b)
    implicit none
    type(batch_dgemm_parameter), intent(inout) :: b
    call dgemm_batch(b%transa, b%transb, b%m, b%n, b%k, b%alpha, b%ap, b%lda, b%bp, b%ldb, &
                     b%beta, b%cp, b%ldc, b%group_count, b%group_size)
  end subroutine
end

module batch_sep_dc_type
  implicit none
  type :: batch_dlaed1ab_parameter
    integer :: idlmda, iw, iq2, indxc, coltyp, k
  endtype
  type :: batch_dlaed2_parameter
    integer, pointer :: k_ptr
    integer :: n, cutpnt
    real(kind(0.0d0)), pointer :: d_ptr, q_ptr
    integer :: ldq
    integer, pointer :: indxq_ptr
    real(kind(0.0d0)), pointer :: rho_ptr
    real(kind(0.0d0)), pointer :: work_iz_ptr, work_idlmda_ptr, work_iw_ptr, work_iq2_ptr
    integer, pointer :: iwork_indx_ptr, iwork_indxc_ptr, iwork_indxp_ptr, iwork_coltyp_ptr
  endtype
end


subroutine batch_sep_dc(n, d, e, x, ldx, work, lwork, iwork, liwork, s2, omp_iwork, info, timer)
  use omp_lib
  use batch_sep_dc_timing
  implicit none
  integer, intent(in) :: n
  real(kind(0.0d0)), intent(inout) :: d(*), e(*)
  real(kind(0.0d0)), intent(out) :: x(ldx, *)
  integer, intent(in) :: ldx
  real(kind(0.0d0)), intent(out) :: work(*)
  integer, intent(out) :: iwork(*)
  integer, intent(in) :: lwork, liwork
  real(kind(0.0d0)), intent(out) :: s2(*)
  integer, intent(out) :: omp_iwork(*)
  integer, intent(out) :: info
  type(batch_sep_dc_timing_list), intent(inout) :: timer
  !
  real(kind(0.0d0)) :: t1, t2

  if (lwork == -1 .or. liwork == -1) then
    call dstedc_batch("I", n, d, e, x, ldx, work, lwork, iwork, liwork, s2, omp_iwork, info, timer)
    return
  endif
  
  t1 = omp_get_wtime()
  call dstedc_batch("I", n, d, e, x, ldx, work, lwork, iwork, liwork, s2, omp_iwork, info, timer)
  t2 = omp_get_wtime()
  timer%t_dstedc_batch = t2 - t1

  return
end

