! LOGSTEP Module: dlaed0b_batch.F90
!
! * This program is modified from LAPACK DLAED0.f by Yusuke HIROTA.
!
! # License
! ## Original portions of DLAED0.f
! Copyright (c) 1992-2013 The University of Tennessee and The University
!                         of Tennessee Research Foundation.  All rights
!                         reserved.
! Copyright (c) 2000-2013 The University of California Berkeley. All
!                         rights reserved.
! Copyright (c) 2006-2013 The University of Colorado Denver.  All rights
!                         reserved.
!
! $COPYRIGHT$
!
! Additional copyrights may follow
!
! $HEADER$
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are
! met:
!
! - Redistributions of source code must retain the above copyright
!   notice, this list of conditions and the following disclaimer.
!
! - Redistributions in binary form must reproduce the above copyright
!   notice, this list of conditions and the following disclaimer listed
!   in this license in the documentation and/or other materials
!   provided with the distribution.
!
! - Neither the name of the copyright holders nor the names of its
!   contributors may be used to endorse or promote products derived from
!   this software without specific prior written permission.
!
! The copyright holders provide no reassurances that the source code
! provided does not infringe any patent, copyright, or any other
! intellectual property rights of third parties.  The copyright holders
! disclaim any liability to any recipient for claims brought against
! recipient by any third party for infringement of that parties
! intellectual property rights.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
! "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
! LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
! A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
! OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
! SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
! LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
! DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
! THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
! (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
! OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!
! ## Portions written by Yusuke HIROTA (MIT License)
! Copyright 2018 RIKEN.
!
! Permission is hereby granted, free of charge, to any person obtaining
! a copy of this software and associated documentation files (the
! "Software"), to deal in the Software without restriction, including
! without limitation the rights to use, copy, modify, merge, publish,
! distribute, sublicense, and/or sell copies of the Software, and to
! permit persons to whom the Software is furnished to do so, subject to
! the following conditions:
!
! The above copyright notice and this permission notice shall be
! included in all copies or substantial portions of the Software.
!
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
! EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
! MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
! IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
! CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
! TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
! SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
!
subroutine dlaed0b_batch(subpbs, indxq, d, e, q, ldq, work, iwork, s2, omp_iwork, info, timer)
  use omp_lib
  use batch_sep_dc_timing
  use batch_dgemm
  use batch_sep_dc_type
  implicit none
  integer, intent(inout) :: subpbs ! Num. subproblems
  integer, intent(in) :: indxq
  real(kind(0.0d0)), intent(inout) :: d(*), e(*)
  real(kind(0.0d0)), intent(inout) :: q(ldq, *)
  integer, intent(in) :: ldq
  real(kind(0.0d0)), intent(inout) :: work(*)
  integer, intent(inout) :: iwork(*)
  real(kind(0.0d0)), intent(out) :: s2(*) ! workspace, n ** 2 or more.
  integer, intent(out) :: omp_iwork(*) ! workspace, n * 4 or more.
  integer, intent(out) :: info
  type(batch_sep_dc_timing_list), intent(inout) :: timer
  !
  real(kind(0.0d0)) :: t1, t2
  integer :: spm2, submat(subpbs / 2), matsiz(subpbs / 2), msd2(subpbs / 2), curprb
  integer :: s2ptr(subpbs / 2)
  integer :: i
  !
  type(batch_dgemm_parameter) :: b
  type(batch_dlaed1ab_parameter) :: v_ab(subpbs / 2)
  type(batch_dlaed2_parameter) :: v2(subpbs / 2)
  double precision, parameter :: one = 1.0d0, zero = 0.0d0

  b%group_count = subpbs
  allocate(b%group_size(b%group_count))
  b%group_size(1 : b%group_count) = 1
  call batch_dgemm_parameter_allocate(b)
  b%transa(:) = 'N'
  b%transb(:) = 'N'
  b%alpha(:) = one
  b%beta(:) = zero

  do while (subpbs > 1) 
    if (subpbs == 2) then
      timer%current_level = 1
    else if(subpbs == 4) then
      timer%current_level = 2
    else if(subpbs == 8) then
      timer%current_level = 3
    else if(subpbs == 16) then
      timer%current_level = 4
    else if(subpbs == 32) then
      timer%current_level = 5
    else
      timer%current_level = 6
    endif
    spm2 = subpbs - 2

    !> Index
    do i = 0, spm2, 2
      curprb = i / 2 ! 0, 1, 2, ...
      if (i == 0) then
        submat(curprb + 1) = 1
        matsiz(curprb + 1) = iwork(2)
        msd2(curprb + 1) = iwork(1)
        s2ptr(curprb + 1) = 1
      else
        s2ptr(curprb + 1) = s2ptr(curprb) + matsiz(curprb) ** 2 ! This must be done before the update of submat.
        submat(curprb + 1) = iwork(i) + 1
        matsiz(curprb + 1) = iwork(i + 2) - iwork(i)
        msd2(curprb + 1) = matsiz(curprb + 1) / 2
      end if
      iwork(i / 2 + 1) = iwork(i + 2)
    enddo

    !> dlaed1a
    t1 = omp_get_wtime()
    !$omp parallel do if (subpbs / 2 > 8)
    do i = 0, spm2, 2
      curprb = i / 2 ! 0, 1, 2, ...
      call dlaed1a_batch(matsiz(curprb + 1), d(submat(curprb + 1)), q(submat(curprb + 1), submat(curprb + 1)), ldq, &
                         iwork(indxq + submat(curprb + 1)), e(submat(curprb + 1) + msd2(curprb + 1) - 1), msd2(curprb + 1), &
                         work((submat(curprb + 1) - 1) * 4 + (submat(curprb + 1) - 1) ** 2 + 1), omp_iwork((submat(curprb + 1) - 1) * 4 + 1), &
                         v_ab(curprb + 1), v2(curprb + 1), info)
    enddo
    !$omp end parallel do
    t2 = omp_get_wtime()
    timer%t_dlaed1a_batch(timer%current_level) = timer%t_dlaed1a_batch(timer%current_level) + t2 - t1

    !> dlaed2_batch
    t1 = omp_get_wtime()
    call dlaed2_batch(v2, subpbs / 2, info)
    if (info /= 0) then
      call abort()
    end if
    t2 = omp_get_wtime()
    timer%t_dlaed2(timer%current_level) = timer%t_dlaed2(timer%current_level) + (t2 - t1)

    t1 = omp_get_wtime()
    !> dlaed1b
    !> dlaed4_batch
    !> dlaed1c
    !$omp parallel do if (subpbs / 2 > 8)
    do i = 0, spm2, 2
      curprb = i / 2 ! 0, 1, 2, ...
      call dlaed1b_batch(matsiz(curprb + 1), d(submat(curprb + 1)), q(submat(curprb + 1), submat(curprb + 1)), ldq, &
                         iwork(indxq + submat(curprb + 1)), e(submat(curprb + 1) + msd2(curprb + 1) - 1), msd2(curprb + 1), &
                         work((submat(curprb + 1) - 1) * 4 + (submat(curprb + 1) - 1) ** 2 + 1), omp_iwork((submat(curprb + 1) - 1) * 4 + 1), &
                         s2(s2ptr(curprb + 1)), b, i + 1, v_ab(curprb + 1), info, timer)
    enddo
    !$omp end parallel do
    t2 = omp_get_wtime()
    timer%t_dlaed1b_batch(timer%current_level) = timer%t_dlaed1b_batch(timer%current_level) + t2 - t1

    !> Batched GEMM
    b%group_count = subpbs
    t1 = omp_get_wtime()
    call batch_dgemm_wrapper(b)
    t2 = omp_get_wtime()
    timer%t_batch_dgemm(timer%current_level) = timer%t_batch_dgemm(timer%current_level) + t2 - t1

    subpbs = subpbs / 2
  end do

  call batch_dgemm_parameter_deallocate(b)
  deallocate(b%group_size)

  return
end

