! LOGSTEP Module: dlaed1a_batch.F90
!
! * This program is modified from LAPACK DLAED1.f by Yusuke HIROTA.
!
! # License
! ## Original portions of DLAED1.f
! Copyright (c) 1992-2013 The University of Tennessee and The University
!                         of Tennessee Research Foundation.  All rights
!                         reserved.
! Copyright (c) 2000-2013 The University of California Berkeley. All
!                         rights reserved.
! Copyright (c) 2006-2013 The University of Colorado Denver.  All rights
!                         reserved.
!
! $COPYRIGHT$
!
! Additional copyrights may follow
!
! $HEADER$
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are
! met:
!
! - Redistributions of source code must retain the above copyright
!   notice, this list of conditions and the following disclaimer.
!
! - Redistributions in binary form must reproduce the above copyright
!   notice, this list of conditions and the following disclaimer listed
!   in this license in the documentation and/or other materials
!   provided with the distribution.
!
! - Neither the name of the copyright holders nor the names of its
!   contributors may be used to endorse or promote products derived from
!   this software without specific prior written permission.
!
! The copyright holders provide no reassurances that the source code
! provided does not infringe any patent, copyright, or any other
! intellectual property rights of third parties.  The copyright holders
! disclaim any liability to any recipient for claims brought against
! recipient by any third party for infringement of that parties
! intellectual property rights.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
! "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
! LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
! A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
! OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
! SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
! LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
! DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
! THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
! (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
! OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!
! ## Portions written by Yusuke HIROTA (MIT License)
! Copyright 2018 RIKEN.
!
! Permission is hereby granted, free of charge, to any person obtaining
! a copy of this software and associated documentation files (the
! "Software"), to deal in the Software without restriction, including
! without limitation the rights to use, copy, modify, merge, publish,
! distribute, sublicense, and/or sell copies of the Software, and to
! permit persons to whom the Software is furnished to do so, subject to
! the following conditions:
!
! The above copyright notice and this permission notice shall be
! included in all copies or substantial portions of the Software.
!
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
! EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
! MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
! IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
! CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
! TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
! SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
!
subroutine dlaed1a_batch(n, d, q, ldq, indxq, rho, cutpnt, work, iwork, v, v2, info)
  use :: omp_lib
  use :: batch_sep_dc_timing
  use :: batch_dgemm
  use :: batch_sep_dc_type
  use :: iso_c_binding
  implicit none
  integer, intent(in) :: cutpnt, ldq, n
  integer, intent(out) :: info
  double precision, intent(in) :: rho
  integer :: indxq(*), iwork(*)
  double precision :: d(*), q(ldq, *), work(*)
  !
  integer :: zpp1, iz, indx, indxp ! Used only in this subroutine.
  type(batch_dlaed1ab_parameter), intent(out) :: v ! Parameter list, refered by dlaed1b_batch
  type(batch_dlaed2_parameter), intent(out) :: v2 

  info = 0
  !
  if(n .lt. 0) then
    info = -1
  else if (ldq .lt. max(1, n)) then
    info = -4
  else if (min(1, n / 2) .gt. cutpnt .or. (n / 2) .lt. cutpnt) then
    info = -7
  end if
  if (info /= 0) then
    call xerbla('dlaed1', -info)
    return
  end if

  ! Quick return if possible
  if (n .eq. 0) return

  ! The following values are integer pointers which indicate
  ! the portion of the workspace
  ! used by a particular array in DLAED2 and DLAED3.
  iz = 1
  v%idlmda = iz + n
  v%iw = v%idlmda + n
  v%iq2 = v%iw + n
  !
  indx = 1
  v%indxc = indx + n
  v%coltyp = v%indxc + n
  indxp = v%coltyp + n

  ! Form the z-vector which consists of the last row of Q_1 and the
  ! first row of Q_2.
  call dcopy(cutpnt, q(cutpnt, 1), ldq, work(iz), 1)
  zpp1 = cutpnt + 1
  call dcopy(n - cutpnt, q(zpp1, zpp1), ldq, work(iz + cutpnt), 1)

  ! Deflate eigenvalues.
  call c_f_pointer(c_loc(v%k), v2%k_ptr)
  v2%n = n
  v2%cutpnt = cutpnt
  call c_f_pointer(c_loc(d(1)), v2%d_ptr)
  call c_f_pointer(c_loc(q(1, 1)), v2%q_ptr)
  v2%ldq = ldq
  call c_f_pointer(c_loc(indxq(1)), v2%indxq_ptr)
  call c_f_pointer(c_loc(rho), v2%rho_ptr)
  call c_f_pointer(c_loc(work(iz)), v2%work_iz_ptr)
  call c_f_pointer(c_loc(work(v%idlmda)), v2%work_idlmda_ptr)
  call c_f_pointer(c_loc(work(v%iw)), v2%work_iw_ptr)
  call c_f_pointer(c_loc(work(v%iq2)), v2%work_iq2_ptr)
  call c_f_pointer(c_loc(iwork(indx)), v2%iwork_indx_ptr)
  call c_f_pointer(c_loc(iwork(v%indxc)), v2%iwork_indxc_ptr)
  call c_f_pointer(c_loc(iwork(indxp)), v2%iwork_indxp_ptr)
  call c_f_pointer(c_loc(iwork(v%coltyp)), v2%iwork_coltyp_ptr)
end

