! LOGSTEP Module: dlaed1b_batch.F90
!
! * This program is modified from LAPACK DLAED1.f by Yusuke HIROTA.
!
! # License
! ## Original portions of DLAED1.f
! Copyright (c) 1992-2013 The University of Tennessee and The University
!                         of Tennessee Research Foundation.  All rights
!                         reserved.
! Copyright (c) 2000-2013 The University of California Berkeley. All
!                         rights reserved.
! Copyright (c) 2006-2013 The University of Colorado Denver.  All rights
!                         reserved.
!
! $COPYRIGHT$
!
! Additional copyrights may follow
!
! $HEADER$
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are
! met:
!
! - Redistributions of source code must retain the above copyright
!   notice, this list of conditions and the following disclaimer.
!
! - Redistributions in binary form must reproduce the above copyright
!   notice, this list of conditions and the following disclaimer listed
!   in this license in the documentation and/or other materials
!   provided with the distribution.
!
! - Neither the name of the copyright holders nor the names of its
!   contributors may be used to endorse or promote products derived from
!   this software without specific prior written permission.
!
! The copyright holders provide no reassurances that the source code
! provided does not infringe any patent, copyright, or any other
! intellectual property rights of third parties.  The copyright holders
! disclaim any liability to any recipient for claims brought against
! recipient by any third party for infringement of that parties
! intellectual property rights.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
! "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
! LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
! A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
! OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
! SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
! LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
! DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
! THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
! (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
! OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!
! ## Portions written by Yusuke HIROTA (MIT License)
! Copyright 2018 RIKEN.
!
! Permission is hereby granted, free of charge, to any person obtaining
! a copy of this software and associated documentation files (the
! "Software"), to deal in the Software without restriction, including
! without limitation the rights to use, copy, modify, merge, publish,
! distribute, sublicense, and/or sell copies of the Software, and to
! permit persons to whom the Software is furnished to do so, subject to
! the following conditions:
!
! The above copyright notice and this permission notice shall be
! included in all copies or substantial portions of the Software.
!
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
! EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
! MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
! IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
! CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
! TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
! SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
!
subroutine dlaed1b_batch(n, d, q, ldq, indxq, rho, cutpnt, work, iwork, s2, b, b_head, v, info, timer)
  use omp_lib
  use batch_sep_dc_timing
  use batch_dgemm
  use batch_sep_dc_type
  implicit none
  type(batch_sep_dc_timing_list), intent(inout) :: timer
  real(kind(0.0d0)) :: t1, t2
  double precision, intent(out) :: s2(*) ! workspace, n-by-n or more
  type(batch_dgemm_parameter), intent(inout) :: b
  integer, intent(in) :: b_head
  !
  integer, intent(in) :: cutpnt, ldq, n
  integer, intent(out) :: info
  double precision, intent(in) :: rho
  integer :: indxq(*), iwork(*)
  double precision :: d(*), q(ldq, *), work(*)
  !
  type(batch_dlaed1ab_parameter), intent(out) :: v ! Parameter list, written by dlaed1a_batch
  !
  integer :: is, n1, n2, i ! Used only in dlaed1b 

  info = 0

  ! Solve Secular Equation.
  if (v%k /= 0) then
    is = (iwork(v%coltyp) + iwork(v%coltyp + 1)) * cutpnt + (iwork(v%coltyp + 1) + iwork(v%coltyp + 2)) * (n - cutpnt) + v%iq2
    t1 = omp_get_wtime()
    call dlaed3_batch(v%k, n, cutpnt, d, q, ldq, rho, work(v%idlmda), work(v%iq2), iwork(v%indxc), iwork(v%coltyp), work(v%iw), work(is), s2, b, b_head, info, timer)
    t2 = omp_get_wtime()
    timer%t_dlaed3_batch(timer%current_level) = timer%t_dlaed3_batch(timer%current_level) + (t2 - t1)
    if (info /= 0) call abort()

    ! Prepare the INDXQ sorting permutation.
    n1 = v%k
    n2 = n - v%k
    call dlamrg(n1, n2, d, 1, -1, indxq)
  else
    do i = 1, n
      indxq(i) = i
    enddo
  end if
endsubroutine

