! LOGSTEP Module: dlaed3_batch.F90
!
! * This program is modified from LAPACK DLAED3.f by Yusuke HIROTA.
!
! # License
! ## Original portions of DLAED3.f
! Copyright (c) 1992-2013 The University of Tennessee and The University
!                         of Tennessee Research Foundation.  All rights
!                         reserved.
! Copyright (c) 2000-2013 The University of California Berkeley. All
!                         rights reserved.
! Copyright (c) 2006-2013 The University of Colorado Denver.  All rights
!                         reserved.
!
! $COPYRIGHT$
!
! Additional copyrights may follow
!
! $HEADER$
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are
! met:
!
! - Redistributions of source code must retain the above copyright
!   notice, this list of conditions and the following disclaimer.
!
! - Redistributions in binary form must reproduce the above copyright
!   notice, this list of conditions and the following disclaimer listed
!   in this license in the documentation and/or other materials
!   provided with the distribution.
!
! - Neither the name of the copyright holders nor the names of its
!   contributors may be used to endorse or promote products derived from
!   this software without specific prior written permission.
!
! The copyright holders provide no reassurances that the source code
! provided does not infringe any patent, copyright, or any other
! intellectual property rights of third parties.  The copyright holders
! disclaim any liability to any recipient for claims brought against
! recipient by any third party for infringement of that parties
! intellectual property rights.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
! "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
! LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
! A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
! OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
! SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
! LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
! DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
! THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
! (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
! OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!
! ## Portions written by Yusuke HIROTA (MIT License)
! Copyright 2018 RIKEN.
!
! Permission is hereby granted, free of charge, to any person obtaining
! a copy of this software and associated documentation files (the
! "Software"), to deal in the Software without restriction, including
! without limitation the rights to use, copy, modify, merge, publish,
! distribute, sublicense, and/or sell copies of the Software, and to
! permit persons to whom the Software is furnished to do so, subject to
! the following conditions:
!
! The above copyright notice and this permission notice shall be
! included in all copies or substantial portions of the Software.
!
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
! EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
! MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
! IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
! CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
! TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
! SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
!
subroutine dlaed3_batch(k, n, n1, d, q, ldq, rho, dlamda, q2, indx, ctot, w, s, s2, b, b_head, info, timer)
  use omp_lib
  use batch_sep_dc_timing
  use batch_dgemm
  real(kind(0.0d0)) :: t1, t2
  type(batch_sep_dc_timing_list) :: timer
!
!  -- LAPACK routine (version 3.2) --
!  -- LAPACK is a software package provided by Univ. of Tennessee,    --
!  -- Univ. of California Berkeley, Univ. of Colorado Denver and NAG Ltd..--
!     November 2006
!
  integer :: info, k, ldq, n, n1
  double precision :: rho
  double precision, intent(out) :: s2(*) ! Used for workspace and eigenvectors of the deflated system, k^2 or more 
  type(batch_dgemm_parameter), intent(inout) :: b
  integer, intent(in) :: b_head ! [b_head : b_head + 1] are used.
!
  integer :: ctot( * ), indx( * )
  double precision :: d( * ), dlamda( * ), q( ldq, * ), q2( * ), s( * ), w( * )
!
! .. Parameters ..
  double precision, parameter :: one = 1.0d0, zero = 0.0d0
! .. Local Scalars ..
  integer :: i, ii, iq2, j, n12, n2, n23
  double precision :: temp
!
! .. External Functions ..
  double precision   dlamc3, dnrm2

  !
  ! Test the input parameters.
  !
  info = 0

  if( k.lt.0 ) then
    info = -1
  else if( n.lt.k ) then
    info = -2
  else if( ldq.lt.max( 1, n ) ) then
    info = -6
  end if
  if( info.ne.0 ) then
    call xerbla( 'dlaed3', -info )
    return
  end if

  ! Quick return if possible
  if ( k.eq.0 ) return

  !$omp parallel do 
  do i = 1, k
    dlamda( i ) = dlamc3( dlamda( i ), dlamda( i ) ) - dlamda( i )
  enddo
  !$omp end parallel do

  t1 = omp_get_wtime()
  !$omp parallel do schedule(guided) private(info)
  do j = 1, k
    call dlaed4( k, j, dlamda, w, q( 1, j ), rho, d( j ), info )
    ! If the zero finder fails, the computation is terminated.
    if( info.ne.0 ) call abort()
  enddo
  !$omp end parallel do
  t2 = omp_get_wtime()
  timer%t_dlaed4(timer%current_level) = timer%t_dlaed4(timer%current_level) + t2 - t1

  if( k.eq.1 ) go to 110
  if( k.eq.2 ) then
    do j = 1, k
      w( 1 ) = q( 1, j )
      w( 2 ) = q( 2, j )
      ii = indx( 1 )
      q( 1, j ) = w( ii )
      ii = indx( 2 )
      q( 2, j ) = w( ii )
    enddo
    go to 110
  end if

  !
  ! Compute updated W.
  !
  call dcopy( k, w, 1, s, 1 )
  ! Initialize W(I) = Q(I,I)
  call dcopy( k, q, ldq+1, w, 1 )
  t1 = omp_get_wtime()
  !$omp parallel do private(j) 
  do i = 1, k
    do j = 1, k
      if (i == j) then
        cycle
      endif
      w( i ) = w( i )*( q( i, j ) / ( dlamda( i )-dlamda( j ) ) )
    enddo
    w( i ) = sign( sqrt( -w( i ) ), s( i ) )
  enddo
  !$omp end parallel do
  t2 = omp_get_wtime()
  timer%t_general(timer%current_level) = timer%t_general(timer%current_level) + t2 - t1

  !
  ! Compute eigenvectors of the modified rank-1 modification.
  !
  t1 = omp_get_wtime()
  !$omp parallel do private(i, temp, ii)
  do j = 1, k
    do i = 1, k
      s2((j - 1) * k + i) = w(i) / q(i, j)
    enddo
    temp = dnrm2(k, s2((j - 1) * k + 1), 1)
    do i = 1, k
      ii = indx( i )
      q(i, j) = s2((j - 1) * k + ii) / temp
    enddo
  enddo
  !$omp end parallel do
  t2 = omp_get_wtime()
  timer%t_general(timer%current_level) = timer%t_general(timer%current_level) + t2 - t1

  !
  ! Compute the updated eigenvectors.
  !
  110 CONTINUE

  N2 = N - N1
  N12 = CTOT( 1 ) + CTOT( 2 )
  N23 = CTOT( 2 ) + CTOT( 3 )

  IQ2 = N1*N12 + 1

  call dlacpy('A', k, k, q, ldq, s2, k)

  ! For n12 == 0 or n23 ==0, store the trivial results.
  if (n12 == 0) then
    call dlaset('A', n1, k, zero, zero, q(1, 1), ldq)
  endif
  if (n23 == 0) then
    call dlaset('A', n2, k, zero, zero, q(n1 + 1, 1), ldq)
  endif

  ! Register DGEMMs.
  ! If n12 == 0 or n23 ==0, one or more GEMM(s) do(es) nothing, thus the trivial results will not be changed.
  ! s2(*) : k-by-k, upper (N12-by-K) and lower (N23-by-K), the upper and lower blocks are overlapped.
  b%m(b_head) = n1
  b%n(b_head) = k
  b%k(b_head) = n12
  b%m(b_head + 1) = n2
  b%n(b_head + 1) = k
  b%k(b_head + 1) = n23
  b%ap(b_head) = LOC(q2)
  b%bp(b_head) = LOC(s2)
  b%cp(b_head) = LOC(q)
  b%ap(b_head + 1) = LOC(q2(iq2))
  b%bp(b_head + 1) = LOC(s2(k + 1 - n23))
  b%cp(b_head + 1) = LOC(q(n1 + 1, 1))
  b%lda(b_head) = n1
  b%lda(b_head + 1) = n2
  b%ldb(b_head : b_head + 1) = k
  b%ldc(b_head : b_head + 1) = ldq

  return
end

